//
//  main.c
//  test_shm
//
// Share memory interface for LabVIEW (OSX), remap shmget(), shmat(), shmdt(), shmctl() and copy shm data to/fram LV
//
//  Created by Chris© on 02.06.2019.
//
// 9.8.2023, v.0.1.3, ChS, Apple Silicon support
// 9.8.2019, v.0.1.2, ChS, return err code + err string -> new param -> new CLN param -> incompatible with previous version !!!
//                          ***** for some reasons, LV crashes if I do not specify the errStr min size in the CLN params ¿?¿, crash even if I wire a str cst ??
//  ~7.2019, v.0.1.1, ChS, added var Shm key ID
//
// from http://www.stuffaboutcode.com/2013/08/shared-memory-c-python-ipc.html
//    	https://github.com/martinohanlon/c_python_ipc
//
// also see Steven's http://poincare.matf.bg.ac.rs/~ivana/courses/ps/sistemi_knjige/pomocno/apue/APUE/0201433079/ch15lev1sec9.html
//
// to compile without XCode
//
//      gcc -o shmwriter shmwriter.c
//
// in the terminal, to see what the system sees
//
//      ipcs        # to see shared mem
//      ipcrm       # to remove a given shared mem segment
//

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h> // for sleep
#include <errno.h>
#include <time.h>
#include <syslog.h>
#if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_12
#include <os/log.h>
#endif


#define toConsole(msg, ...)                        \
do {                                \
char i[512], s[512], f[1024];  \
sprintf(i,"[%s(), l.%d] ",__func__, __LINE__); \
sprintf(s,msg, ##__VA_ARGS__); \
sprintf(f,"%s%s",i,s);   \
/*printf(f);*/    \
os_log(OS_LOG_DEFAULT, "%s",(f)); \
} while(0)


#define TmpErrStrSz 1024

#define kLVShmerr_fatal -666
#define kLVShmerr_shmget -77001
#define kLVShmerr_shmat -77002
#define kLVShmerr_shmdt -77003
#define kLVShmerr_shmctl -77004
#define kLVShmerr_shmatToManyFileOpen -77024



// --------------------------------------------------------------------------
int32_t LVShm_Version(char *VersionStr) {
#ifdef __LP64__
    sprintf(VersionStr,"LV_Shm v.0.1.2 - 64b (%s)",__TIMESTAMP__);
#else
    sprintf(VersionStr,"LV_Shm v.0.1.2 - 32b (%s)",__TIMESTAMP__);
#endif
    
    return 0;
}


// --------------------------------------------------------------------------
int LVShm_Create(size_t size, key_t shmKey, int *shmid, uintptr_t *shared_memoryPtr,char *errStr){ // was char **shared_memoryPtr
    
    int err = 0;
    char tmp[TmpErrStrSz];
    strcpy(errStr,"");
    
    sprintf(tmp,"sizeof(int): %ld, sizeof(long): %ld, sizeof(uint64_t): %ld",sizeof(int),sizeof(long),sizeof(uint64_t));
    
    // key_t is int32 (for now)
    // Setup shared memory
    if ((*shmid = shmget(shmKey, size, IPC_CREAT | 0666)) < 0)
    {
        err = kLVShmerr_shmget;
        sprintf(tmp,"LVShm_Create() - Error (%d) getting shared memory id for key 0x%x, size: %zu, errno: %d, '%s'",err, shmKey,size,errno,strerror(errno));
        //sprintf(tmp,"LVShm_Create() - Error ");
        stpncpy(errStr,tmp,TmpErrStrSz);
        *shmid=-1; //what is a good value in case of problem ? 0? -1 ?
        goto fail;
    }
    // Attached shared memory
    if ((*shared_memoryPtr = shmat(*shmid, NULL, 0)) == (uintptr_t) -1)
    {
        if (errno==24) // UGLY,
            err = kLVShmerr_shmatToManyFileOpen;
        else
            err = kLVShmerr_shmat;
        sprintf(tmp,"LVShm_Create() - Error (%d) attaching shared memory id for key 0x%x, size: %zu, errno: %d,'%s'",err,shmKey,size,errno,strerror(errno));
        //sprintf(tmp,"LVShm_Create() - Error 2");
        stpncpy(errStr,tmp,TmpErrStrSz);
        *shmid=-1; //what is a good value in case of problem ? 0? -1 ?
        // *shared_memoryPtr = NULL;
        // shmctl((int)*shmid, IPC_RMID, NULL);  // should do it here ?
        goto fail;
    }
    
    sprintf(tmp,"LVShm_Create(Key:%d, sz:%zu) -> id: %d - done",shmKey,size,*shmid);
    //sprintf(tmp,"LVShm_Create - done");
    
    stpncpy(errStr,tmp,TmpErrStrSz);
    
fail:
    return err;
}

// --------------------------------------------------------------------------
int LVShm_Set(char *shared_memoryPtr, char *data, int size){
    memcpy(shared_memoryPtr, data, size); // test if shared_memory is NULL and if size > size(shared_memory)
    return 0; // fix err handling
}

// --------------------------------------------------------------------------
int LVShm_Get(char *shared_memoryPtr, char *data, int size){
    memcpy(data,shared_memoryPtr, size); // test if shared_memory is NULL and if size > size(shared_memory)
    return 0; // fix err handling
}

// --------------------------------------------------------------------------

int LVShm_Clear(int LV_shmid, uintptr_t addr,char *errStr) {

    int err=0   ;
    char tmp[TmpErrStrSz];
    strcpy(errStr,"");

    err=shmdt((const void *)addr); // fix fmt
    if (err !=0)
    {
        err = kLVShmerr_shmdt;
        sprintf(tmp,"LVShm_Clear() - Error (%d) dettaching shared memory id (%d), errno: %d, '%s'",err,LV_shmid, errno,strerror(errno));
        stpncpy(errStr,tmp,TmpErrStrSz);
    }
    err=shmctl((int)LV_shmid, IPC_RMID, NULL);
    if (err !=0)
    {
        err = kLVShmerr_shmctl;
        sprintf(tmp,"LVShm_Clear() - Error (%d) dettaching shared memory id (%d), errno: %d, '%s'",err,LV_shmid, errno,strerror(errno));
        stpncpy(errStr,tmp,TmpErrStrSz);
    }

    sprintf(tmp,"LVShm_Clear(id: %d) - done",LV_shmid);
    stpncpy(errStr,tmp,TmpErrStrSz);
fail:
    return err; // fix err handling

}



// --------------------------------------------------------------------------
#define LVShm_Key  123456
#define LVShm_testStr "Hello Python from LabVIEW 20xx"

int main(int argc, const char **argv)
{
    int err=0;
    int LV_shmid;
    char *LV_shared_memory;
    char *errStr="---------------------------------------------------------------------";
    
    err =LVShm_Create(sizeof(LVShm_testStr),LVShm_Key,&LV_shmid,&LV_shared_memory,errStr);
    //err =LVShm_Create(sizeof(LVShm_testStr),LVShm_Key,&LV_shmid,&LV_shared_memory);
    
    char myStr[10] = "12345678";
    myStr[2]=0;
    
    err =LVShm_Set(LV_shared_memory, myStr, sizeof(LVShm_testStr));
    
    
    // sleep so there is enough time to run the reader!
    time_t curtime;
    time(&curtime);
    printf("start & sleeping, %s", ctime(&curtime));
    printf("press RETURN to wake me up");
    //sleep(30); //Sec
    getchar();
    time(&curtime);
    printf("stop sleeping %s", ctime(&curtime));
    // Detach and remove shared memory
    printf("read '%s'\n", LV_shared_memory);
    
    err=LVShm_Clear(LV_shmid,NULL,errStr);
    
    return err;
}
